from datetime import datetime
from typing import Optional


class History:
    def __init__(
        self,
        first_committed_at: datetime,
        merged_at: datetime,
        deployed_at: Optional[datetime],
    ):
        self.first_committed_at = first_committed_at
        self.merged_at = merged_at
        self.deployed_at = deployed_at

    @staticmethod
    def keys() -> list[str]:
        return [
            "first_committed_at",
            "merged_at",
            "deployed_at",
        ]

    def values(self) -> list:
        return [self.first_committed_at, self.merged_at, self.deployed_at]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, History):
            return NotImplemented
        return (
            self.first_committed_at == other.first_committed_at
            and self.merged_at == other.merged_at
            and self.deployed_at == other.deployed_at
        )


class RepositoryHistory:
    def __init__(
        self, first_committed_at: datetime, merged_at: datetime, merge_commit_sha: str
    ):
        self.first_committed_at = first_committed_at
        self.merged_at = merged_at
        self.merge_commit_sha = merge_commit_sha


class DeploymentHistory:
    def __init__(
        self,
        deployed_at: datetime,
        sha: str,
    ):
        self.deployed_at = deployed_at
        self.sha = sha


def create_histories(
    repository_histories: list[RepositoryHistory],
    deployment_histories: list[DeploymentHistory],
) -> list[History]:
    shaToDeploymentHistoryMap: dict[str, DeploymentHistory] = {}
    for dh in deployment_histories:
        shaToDeploymentHistoryMap[dh.sha] = dh

    histories = []
    for rh in repository_histories:
        matched_dh = shaToDeploymentHistoryMap.get(rh.merge_commit_sha)
        histories.append(
            History(
                rh.first_committed_at,
                rh.merged_at,
                matched_dh.deployed_at if matched_dh is not None else None,
            )
        )

    return histories



import typer
from rich import print

from depwatch.command import generate_histories


def main(
    name: str,
    code_only: bool = typer.Option(
        False, help="do not retrieve deployment histories from CI"
    ),
    limit: int = typer.Option(100, help="count limit for retrieving history"),
):
    generate_histories(name, code_only, limit)
    print(":sparkles::sparkles: [green]Completed![/green] :sparkles::sparkles:")


def main_cli():
    typer.run(main)


if __name__ == "__main__":
    typer.run(main)



from datetime import datetime, timezone
import secrets
from unittest.mock import patch, Mock, MagicMock
from depwatch import repository
from depwatch.exception import DepwatchException


class TestRepository:
    @patch("depwatch.repository.Github")
    def test_get_main_branch_with_main_branch(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_branches(["main"])
        )

        assert repository.get_main_branch("hamakou108/my_project") == "main"

    @patch("depwatch.repository.Github")
    def test_get_main_branch_with_master_branch(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_branches(["master"])
        )

        assert repository.get_main_branch("hamakou108/my_project") == "master"

    @patch("depwatch.repository.Github")
    def test_get_main_branch_with_other_branch(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_branches(["foo"])
        )

        try:
            repository.get_main_branch("hamakou108/my_project")
            assert False
        except DepwatchException as e:
            assert str(e) == "'main' or 'master' branch was not found"

    @patch("depwatch.repository.Github")
    def test_get_repository_history(self, mock_Github: Mock):
        mock_Github.return_value.get_repo.return_value = (
            self.create_mock_repo_with_pulls()
        )

        result = repository.get_repository_history("hamakou108/my_project", "main", 100)

        assert len(result) == 2
        assert result[0].first_committed_at == datetime(2021, 1, 1, tzinfo=timezone.utc)
        assert result[0].merged_at == datetime(2021, 1, 3, tzinfo=timezone.utc)
        assert len(result[0].merge_commit_sha) == 40

    def create_mock_repo_with_branches(self, branches: list[str]):
        mock_repo = MagicMock()

        mock_branches = []
        for b in branches:
            mock_branch = MagicMock(spec=["name"])
            mock_branch.name = b
            mock_branches.append(mock_branch)

        mock_repo.get_branches.return_value = mock_branches

        return mock_repo

    def create_mock_repo_with_pulls(self):
        mock_repo = MagicMock()

        mock_repo.get_pulls.return_value = [
            self.create_mock_pull(True),
            self.create_mock_pull(False),
            self.create_mock_pull(True),
        ]

        return mock_repo

    def create_mock_pull(self, is_merged: bool):
        mock_pull = MagicMock(spec=["get_commits", "merged_at", "merge_commit_sha"])
        mock_commits = []
        for i in range(3):
            mock_commit = MagicMock(spec=["commit"])
            mock_commit.commit.author.date = datetime(
                2021, 1, i + 1, tzinfo=timezone.utc
            )
            mock_commits.append(mock_commit)
        mock_pull.get_commits.return_value = mock_commits
        mock_pull.merged_at = (
            datetime(2021, 1, 3, tzinfo=timezone.utc) if is_merged else None
        )
        mock_pull.merge_commit_sha = secrets.token_hex(20) if is_merged else None

        return mock_pull



from dotenv import load_dotenv

from depwatch.history import create_histories
from depwatch.repository import get_main_branch, get_repository_history
from depwatch.deployment import get_deployment_history
from depwatch.writer import write_histories


def generate_histories(name: str, code_only: bool, limit: int):
    load_dotenv()

    base = get_main_branch(name)

    repository_histories = get_repository_history(name, base, limit)

    # CircleCI
    deployment_histories = []
    if not code_only:
        deployment_histories = get_deployment_history(name, base, limit)

    write_histories(
        "output.csv", create_histories(repository_histories, deployment_histories)
    )



from unittest.mock import patch, Mock

from depwatch import command
from depwatch.command import generate_histories


class TestCommand:
    @patch.object(command, "write_histories")
    @patch.object(command, "get_deployment_history", return_value=[])
    @patch.object(command, "get_repository_history", return_value=[])
    @patch.object(command, "get_main_branch", return_value="main")
    def test_generate_histories(
        self,
        get_main_branch_mock: Mock,
        get_repository_history_mock: Mock,
        get_deployment_history_mock: Mock,
        write_histories_mock: Mock,
    ):
        generate_histories("hamakou108/my_project", False, 100)

        get_main_branch_mock.assert_called_once_with("hamakou108/my_project")
        get_repository_history_mock.assert_called_once_with(
            "hamakou108/my_project", "main", 100
        )
        get_deployment_history_mock.assert_called_once_with(
            "hamakou108/my_project", "main", 100
        )
        write_histories_mock.assert_called()

    @patch.object(command, "write_histories")
    @patch.object(command, "get_deployment_history", return_value=[])
    @patch.object(command, "get_repository_history", return_value=[])
    @patch.object(command, "get_main_branch", return_value="main")
    def test_generate_histories_with_code_only(
        self,
        get_main_branch_mock: Mock,
        get_repository_history_mock: Mock,
        get_deployment_history_mock: Mock,
        write_histories_mock: Mock,
    ):
        generate_histories("hamakou108/my_project", True, 100)

        get_deployment_history_mock.assert_not_called()



import os
from datetime import datetime
from pycircleci.api import Api
from requests import HTTPError
from requests.models import Response
from typing import cast

from depwatch.history import DeploymentHistory


def get_deployment_history(name: str, base: str, limit: int) -> list[DeploymentHistory]:
    def __get_pipeline_workflow(p):
        try:
            return ci.get_pipeline_workflow(p.get("id"))
        except HTTPError as e:
            if cast(Response, e.response).status_code == 404:
                return []
            else:
                raise e

    histories = []

    ci = Api(
        token=os.environ.get("CIRCLECI_ACCESS_TOKEN"), url="https://circleci.com/api"
    )
    user_name, project = name.split("/")
    pipelines = ci.get_project_pipelines(
        user_name, project, branch=base, paginate=True, limit=limit
    )

    for p in pipelines:
        if len(p.get("errors")) != 0:
            continue

        workflows = __get_pipeline_workflow(p)

        stopped_at_list = [w.get("stopped_at") for w in workflows]
        latest_stopped_at = None
        for s in stopped_at_list:
            if s is None:
                continue

            current_stopped_at = datetime.fromisoformat(s)
            if latest_stopped_at is None or current_stopped_at < latest_stopped_at:
                latest_stopped_at = current_stopped_at

        if latest_stopped_at is None:
            continue

        histories.append(
            DeploymentHistory(latest_stopped_at, p.get("vcs").get("revision"))
        )

    return histories



from datetime import datetime, timezone
from unittest.mock import patch, Mock, MagicMock

from requests import HTTPError, Response
from depwatch import deployment

utc = timezone.utc


class TestDeployment:
    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_returns_correct_result(self, mock_Api: Mock):
        mock_ci = MagicMock()
        mock_ci.get_project_pipelines.return_value = [
            {
                "id": "41a70637-129d-4b4c-b285-5a3576470416",
                "errors": [],
                "vcs": {"revision": "c76089814e0bb532ff0634634c4f08aeab4692f3"},
            },
            {
                "id": "da6b8e0d-b3f2-49f4-91d1-9031426f9213",
                "errors": [],
                "vcs": {"revision": "ee2eeb43810fbb1232f28a82af63ad033314bf2b"},
            },
        ]
        mock_ci.get_pipeline_workflow.side_effect = [
            [
                {"stopped_at": "2023-01-01T00:00:00.000Z"},
                {"stopped_at": "2023-01-02T00:00:00.000Z"},
            ],
            [
                {"stopped_at": "2023-01-03T00:00:00.000Z"},
                {"stopped_at": "2023-01-04T00:00:00.000Z"},
            ],
        ]
        mock_Api.return_value = mock_ci

        result = deployment.get_deployment_history("hamakou108/my_project", "main", 100)

        assert len(result) == 2
        assert result[0].deployed_at == datetime(2023, 1, 1, tzinfo=timezone.utc)
        assert result[0].sha == "c76089814e0bb532ff0634634c4f08aeab4692f3"
        assert result[1].deployed_at == datetime(2023, 1, 3, tzinfo=timezone.utc)
        assert result[1].sha == "ee2eeb43810fbb1232f28a82af63ad033314bf2b"

    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_skips_pipelines_with_errors(self, mock_Api: Mock):
        mock_ci = MagicMock()
        mock_ci.get_project_pipelines.return_value = [
            {
                "id": "41a70637-129d-4b4c-b285-5a3576470416",
                "errors": ["error"],
                "vcs": {"revision": "c76089814e0bb532ff0634634c4f08aeab4692f3"},
            }
        ]
        mock_Api.return_value = mock_ci

        result = deployment.get_deployment_history("hamakou108/my_project", "main", 100)

        assert len(result) == 0

    # Access to pipelines that are over six months old is not available on CircleCI
    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_skips_pipelines_with_pipeline_has_removed(
        self, mock_Api: Mock
    ):
        mock_ci = MagicMock()
        mock_ci.get_project_pipelines.return_value = [
            {
                "id": "41a70637-129d-4b4c-b285-5a3576470416",
                "errors": [],
                "vcs": {"revision": "c76089814e0bb532ff0634634c4f08aeab4692f3"},
            }
        ]
        response_404 = Response()
        response_404.status_code = 404
        mock_ci.get_pipeline_workflow.side_effect = HTTPError(response=response_404)
        mock_Api.return_value = mock_ci

        result = deployment.get_deployment_history("hamakou108/my_project", "main", 100)

        assert len(result) == 0

    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_skips_pipelines_with_no_workflows(
        self, mock_Api: Mock
    ):
        mock_ci = MagicMock()
        mock_ci.get_project_pipelines.return_value = [
            {
                "id": "41a70637-129d-4b4c-b285-5a3576470416",
                "errors": [],
                "vcs": {"revision": "c76089814e0bb532ff0634634c4f08aeab4692f3"},
            },
        ]
        mock_ci.get_pipeline_workflow.return_value = []
        mock_Api.return_value = mock_ci

        result = deployment.get_deployment_history("hamakou108/my_project", "main", 100)

        assert len(result) == 0

    @patch("depwatch.deployment.Api")
    def test_get_deployment_history_skips_pipelines_with_no_completed_workflows(
        self, mock_Api: Mock
    ):
        mock_ci = MagicMock()
        mock_ci.get_project_pipelines.return_value = [
            {
                "id": "41a70637-129d-4b4c-b285-5a3576470416",
                "errors": [],
                "vcs": {"revision": "c76089814e0bb532ff0634634c4f08aeab4692f3"},
            },
        ]
        mock_ci.get_pipeline_workflow.return_value = [{"stopped_at": None}]
        mock_Api.return_value = mock_ci

        result = deployment.get_deployment_history("hamakou108/my_project", "main", 100)

        assert len(result) == 0



from datetime import datetime
import os
import shutil
from unittest import TestCase
from unittest.mock import patch
from depwatch.writer import write_histories
from depwatch.history import History


class TestWriter(TestCase):
    test_out_dir = "test_out"

    def setUp(self):
        os.mkdir(TestWriter.test_out_dir)

    def test_write_histories(self):
        histories = [
            History(
                datetime.fromisoformat("2023-02-25T20:00:00+00:00"),
                datetime.fromisoformat("2023-02-26T22:00:00+00:00"),
                datetime.fromisoformat("2023-02-27T10:15:00+00:00"),
            ),
            History(
                datetime.fromisoformat("2023-02-23T10:05:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:15:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:30:00+00:00"),
            ),
            History(
                datetime.fromisoformat("2023-02-23T10:00:00+00:00"),
                datetime.fromisoformat("2023-02-23T11:00:00+00:00"),
                None,
            ),
        ]

        expected_csv = """first_committed_at,merged_at,deployed_at
2023-02-25T20:00:00+00:00,2023-02-26T22:00:00+00:00,2023-02-27T10:15:00+00:00
2023-02-23T10:05:00+00:00,2023-02-23T11:15:00+00:00,2023-02-23T11:30:00+00:00
2023-02-23T10:00:00+00:00,2023-02-23T11:00:00+00:00,
"""

        with patch.dict("os.environ", {"OUTPUT_DIR": TestWriter.test_out_dir}):
            write_histories("test.csv", histories)

        assert os.path.exists(f"{TestWriter.test_out_dir}/test.csv")

        # CSVファイルの内容を確認
        with open(f"{TestWriter.test_out_dir}/test.csv", "r") as f:
            result_csv = f.read()
            assert result_csv == expected_csv

    def tearDown(self):
        shutil.rmtree(TestWriter.test_out_dir)
